@@ -157,7 +157,7 @@ static int uxst_find_compatible_fd(struct listener *l)
 		if (xfer_sock->prev)
 			xfer_sock->prev->next = xfer_sock->next;
 		if (xfer_sock->next)
-			xfer_sock->next->prev = xfer_sock->next->prev;
+			xfer_sock->next->prev = xfer_sock->prev;
 		free(xfer_sock);
 	}
 	return ret;
@@ -396,7 +396,9 @@ static int uxst_pause_listener(struct listener *l)
 	if (((struct sockaddr_un *)&l->addr)->sun_path[0])
 		return 1;
 
-	unbind_listener(l);
+	/* Listener's lock already held. Call lockless version of
+	 * unbind_listener. */
+	do_unbind_listener(l, 1);
 	return 0;
 }
 
